# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from pathlib import PureWindowsPath

from parameterized import parameterized
from twisted.internet import defer
from twisted.internet import error
from twisted.python.reflect import namedModule
from twisted.trial import unittest

from buildbot import config
from buildbot.interfaces import WorkerSetupError
from buildbot.process import buildstep
from buildbot.process import remotetransfer
from buildbot.process.results import FAILURE
from buildbot.process.results import RETRY
from buildbot.process.results import SUCCESS
from buildbot.steps.source import svn
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import ExpectCpdir
from buildbot.test.steps import ExpectDownloadFile
from buildbot.test.steps import ExpectRemoteRef
from buildbot.test.steps import ExpectRmdir
from buildbot.test.steps import ExpectShell
from buildbot.test.steps import ExpectStat
from buildbot.test.util import sourcesteps
from buildbot.test.util.properties import ConstantRenderable


class TestSVN(sourcesteps.SourceStepMixin, TestReactorMixin, unittest.TestCase):
    svn_st_xml = """<?xml version="1.0"?>
        <status>
            <target path=".">
                <entry path="svn_external_path">
                    <wc-status props="none" item="external">
                    </wc-status>
                </entry>
                <entry path="svn_external_path/unversioned_file1">
                    <wc-status props="none" item="unversioned">
                    </wc-status>
                </entry>
                <entry path="svn_external_path/unversioned_file2_uniçode">
                    <wc-status props="none" item="unversioned">
                    </wc-status>
                </entry>
            </target>
        </status>
        """
    svn_st_xml_corrupt = """<?xml version="1.0"?>
            <target path=".">
                <entry path="svn_external_path">
                    <wc-status props="none" item="external">
                    </wc-status>
                </entry>
                <entry path="svn_external_path/unversioned_file">
                    <wc-status props="none" item="unversioned">
                    </wc-status>
                </entry>
            </target>
        </status>
        """
    svn_st_xml_empty = """<?xml version="1.0"?>
                          <status>
                          <target path=".">
                          </target>
                          </status>"""
    svn_info_stdout_xml = """<?xml version="1.0"?>
                            <info>
                            <entry
                               kind="dir"
                               path="."
                               revision="100">
                            <url>http://svn.red-bean.com/repos/test</url>
                            <repository>
                            <root>http://svn.red-bean.com/repos/test</root>
                            <uuid>5e7d134a-54fb-0310-bd04-b611643e5c25</uuid>
                            </repository>
                            <wc-info>
                            <schedule>normal</schedule>
                            <depth>infinity</depth>
                            </wc-info>
                            <commit
                               revision="90">
                            <author>sally</author>
                            <date>2003-01-15T23:35:12.847647Z</date>
                            </commit>
                            </entry>
                            </info>"""
    svn_info_stdout_xml_nonintegerrevision = """<?xml version="1.0"?>
                            <info>
                            <entry
                               kind="dir"
                               path="."
                               revision="a10">
                            <url>http://svn.red-bean.com/repos/test</url>
                            <repository>
                            <root>http://svn.red-bean.com/repos/test</root>
                            <uuid>5e7d134a-54fb-0310-bd04-b611643e5c25</uuid>
                            </repository>
                            <wc-info>
                            <schedule>normal</schedule>
                            <depth>infinity</depth>
                            </wc-info>
                            <commit
                               revision="a10">
                            <author>sally</author>
                            <date>2003-01-15T23:35:12.847647Z</date>
                            </commit>
                            </entry>
                            </info>"""

    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def patch_workerVersionIsOlderThan(self, result):
        self.patch(svn.SVN, 'workerVersionIsOlderThan', lambda x, y, z: result)

    def test_no_repourl(self):
        with self.assertRaises(config.ConfigErrors):
            svn.SVN()

    def test_incorrect_mode(self):
        with self.assertRaises(config.ConfigErrors):
            svn.SVN(repourl='http://svn.local/app/trunk', mode='invalid')

    def test_incorrect_method(self):
        with self.assertRaises(config.ConfigErrors):
            svn.SVN(repourl='http://svn.local/app/trunk', method='invalid')

    def test_svn_not_installed(self):
        self.setup_step(svn.SVN(repourl='http://svn.local/app/trunk'))
        self.expect_commands(ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(1))
        self.expect_exception(WorkerSetupError)
        return self.run_step()

    def test_corrupt_xml(self):
        self.setup_step(svn.SVN(repourl='http://svn.local/app/trunk'))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['svn', 'info', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='wkdir', command=['svn', 'update', '--non-interactive', '--no-auth-cache']
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_st_xml_corrupt)
            .exit(0),
        )
        self.expect_outcome(result=FAILURE)
        return self.run_step()

    @defer.inlineCallbacks
    def test_revision_noninteger(self):
        svnTestStep = svn.SVN(repourl='http://svn.local/app/trunk')
        self.setup_step(svnTestStep)
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['svn', 'info', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='wkdir', command=['svn', 'update', '--non-interactive', '--no-auth-cache']
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml_nonintegerrevision)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', 'a10', 'SVN')
        yield self.run_step()

        revision = self.get_nth_step(0).getProperty('got_revision')
        with self.assertRaises(ValueError):
            int(revision)

    def test_revision_missing(self):
        """Fail if 'revision' tag isn't there"""
        svn_info_stdout = self.svn_info_stdout_xml.replace('entry', 'Blah')

        svnTestStep = svn.SVN(repourl='http://svn.local/app/trunk')
        self.setup_step(svnTestStep)
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['svn', 'info', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='wkdir', command=['svn', 'update', '--non-interactive', '--no-auth-cache']
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(svn_info_stdout)
            .exit(0),
        )
        self.expect_outcome(result=FAILURE)
        return self.run_step()

    def test_mode_incremental(self):
        self.setup_step(
            svn.SVN(
                repourl='http://svn.local/app/trunk',
                mode='incremental',
                username='user',
                password='pass',
                extra_args=['--random'],
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'info',
                    '--xml',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--username',
                    'user',
                    '--password',
                    ('obfuscated', 'pass', 'XXXXXX'),
                    '--random',
                ],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'update',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--username',
                    'user',
                    '--password',
                    ('obfuscated', 'pass', 'XXXXXX'),
                    '--random',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_incremental_timeout(self):
        self.setup_step(
            svn.SVN(
                repourl='http://svn.local/app/trunk',
                mode='incremental',
                username='user',
                timeout=1,
                password='pass',
                extra_args=['--random'],
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', timeout=1, command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                timeout=1,
                command=[
                    'svn',
                    'info',
                    '--xml',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--username',
                    'user',
                    '--password',
                    ('obfuscated', 'pass', 'XXXXXX'),
                    '--random',
                ],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                timeout=1,
                command=[
                    'svn',
                    'update',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--username',
                    'user',
                    '--password',
                    ('obfuscated', 'pass', 'XXXXXX'),
                    '--random',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', timeout=1, command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_incremental_repourl_renderable(self):
        self.setup_step(
            svn.SVN(repourl=ConstantRenderable('http://svn.local/trunk'), mode='incremental')
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['svn', 'info', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout("""<?xml version="1.0"?>""" + """<url>http://svn.local/trunk</url>""")
            .exit(0),
            ExpectShell(
                workdir='wkdir', command=['svn', 'update', '--non-interactive', '--no-auth-cache']
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_incremental_repourl_canonical(self):
        self.setup_step(svn.SVN(repourl='http://svn.local/trunk/test app', mode='incremental'))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['svn', 'info', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/trunk/test%20app</url>')
            .exit(0),
            ExpectShell(
                workdir='wkdir', command=['svn', 'update', '--non-interactive', '--no-auth-cache']
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_incremental_repourl_not_updatable(self):
        self.setup_step(
            svn.SVN(
                repourl=ConstantRenderable('http://svn.local/trunk/app'),
                mode='incremental',
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'checkout',
                    'http://svn.local/trunk/app',
                    '.',
                    '--non-interactive',
                    '--no-auth-cache',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_incremental_retry(self):
        self.setup_step(
            svn.SVN(
                repourl=ConstantRenderable('http://svn.local/trunk/app'),
                mode='incremental',
                retry=(0, 1),
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'checkout',
                    'http://svn.local/trunk/app',
                    '.',
                    '--non-interactive',
                    '--no-auth-cache',
                ],
            ).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'checkout',
                    'http://svn.local/trunk/app',
                    '.',
                    '--non-interactive',
                    '--no-auth-cache',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_incremental_repourl_not_updatable_svninfo_mismatch(self):
        self.setup_step(
            svn.SVN(repourl=ConstantRenderable('http://svn.local/trunk/app'), mode='incremental')
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['svn', 'info', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            # expecting ../trunk/app
            .stdout('<?xml version="1.0"?><url>http://svn.local/branch/foo/app</url>')
            .exit(0),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'checkout',
                    'http://svn.local/trunk/app',
                    '.',
                    '--non-interactive',
                    '--no-auth-cache',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_incremental_given_revision(self):
        self.setup_step(
            svn.SVN(repourl='http://svn.local/app/trunk', mode='incremental'), {"revision": '100'}
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['svn', 'info', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'update',
                    '--revision',
                    '100',
                    '--non-interactive',
                    '--no-auth-cache',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_incremental_win32path(self):
        self.setup_step(
            svn.SVN(
                repourl='http://svn.local/app/trunk',
                mode='incremental',
                username='user',
                password='pass',
                extra_args=['--random'],
            )
        )
        self.build.path_module = namedModule("ntpath")
        self.build.path_cls = PureWindowsPath
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file=r'wkdir\.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file=r'wkdir\.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'info',
                    '--xml',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--username',
                    'user',
                    '--password',
                    ('obfuscated', 'pass', 'XXXXXX'),
                    '--random',
                ],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'update',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--username',
                    'user',
                    '--password',
                    ('obfuscated', 'pass', 'XXXXXX'),
                    '--random',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_incremental_preferLastChangedRev(self):
        """Give the last-changed rev if 'preferLastChangedRev' is set"""
        self.setup_step(
            svn.SVN(
                repourl='http://svn.local/app/trunk',
                mode='incremental',
                username='user',
                preferLastChangedRev=True,
                password='pass',
                extra_args=['--random'],
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'info',
                    '--xml',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--username',
                    'user',
                    '--password',
                    ('obfuscated', 'pass', 'XXXXXX'),
                    '--random',
                ],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'update',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--username',
                    'user',
                    '--password',
                    ('obfuscated', 'pass', 'XXXXXX'),
                    '--random',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '90', 'SVN')
        return self.run_step()

    def test_mode_incremental_preferLastChangedRev_butMissing(self):
        """If 'preferLastChangedRev' is set, but missing, fall back
        to the regular revision value."""
        svn_info_stdout = self.svn_info_stdout_xml.replace('commit', 'Blah')

        self.setup_step(
            svn.SVN(
                repourl='http://svn.local/app/trunk',
                mode='incremental',
                username='user',
                preferLastChangedRev=True,
                password='pass',
                extra_args=['--random'],
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'info',
                    '--xml',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--username',
                    'user',
                    '--password',
                    ('obfuscated', 'pass', 'XXXXXX'),
                    '--random',
                ],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'update',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--username',
                    'user',
                    '--password',
                    ('obfuscated', 'pass', 'XXXXXX'),
                    '--random',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(svn_info_stdout)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_full_clobber(self):
        self.setup_step(
            svn.SVN(repourl='http://svn.local/app/trunk', mode='full', method='clobber')
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'checkout',
                    'http://svn.local/app/trunk',
                    '.',
                    '--non-interactive',
                    '--no-auth-cache',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_full_clobber_given_revision(self):
        self.setup_step(
            svn.SVN(repourl='http://svn.local/app/trunk', mode='full', method='clobber'),
            {"revision": '100'},
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'checkout',
                    'http://svn.local/app/trunk',
                    '.',
                    '--revision',
                    '100',
                    '--non-interactive',
                    '--no-auth-cache',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_full_fresh(self):
        self.setup_step(
            svn.SVN(
                repourl='http://svn.local/app/trunk', mode='full', method='fresh', depth='infinite'
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'info',
                    '--xml',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--depth',
                    'infinite',
                ],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'status',
                    '--xml',
                    '--no-ignore',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--depth',
                    'infinite',
                ],
            )
            .stdout(self.svn_st_xml_empty)
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'update',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--depth',
                    'infinite',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .stdout('\n')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_full_fresh_retry(self):
        self.setup_step(
            svn.SVN(repourl='http://svn.local/app/trunk', mode='full', method='fresh', retry=(0, 2))
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'checkout',
                    'http://svn.local/app/trunk',
                    '.',
                    '--non-interactive',
                    '--no-auth-cache',
                ],
            ).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'checkout',
                    'http://svn.local/app/trunk',
                    '.',
                    '--non-interactive',
                    '--no-auth-cache',
                ],
            ).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'checkout',
                    'http://svn.local/app/trunk',
                    '.',
                    '--non-interactive',
                    '--no-auth-cache',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .stdout('\n')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_full_fresh_given_revision(self):
        self.setup_step(
            svn.SVN(
                repourl='http://svn.local/app/trunk', mode='full', method='fresh', depth='infinite'
            ),
            {"revision": '100'},
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'info',
                    '--xml',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--depth',
                    'infinite',
                ],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'status',
                    '--xml',
                    '--no-ignore',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--depth',
                    'infinite',
                ],
            )
            .stdout(self.svn_st_xml_empty)
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'update',
                    '--revision',
                    '100',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--depth',
                    'infinite',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .stdout('\n')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_full_fresh_keep_on_purge(self):
        self.setup_step(
            svn.SVN(
                repourl='http://svn.local/app/trunk',
                mode='full',
                keep_on_purge=['svn_external_path/unversioned_file1'],
            )
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['svn', 'info', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'status',
                    '--xml',
                    '--no-ignore',
                    '--non-interactive',
                    '--no-auth-cache',
                ],
            )
            .stdout(self.svn_st_xml)
            .exit(0),
            ExpectRmdir(
                dir=['wkdir/svn_external_path/unversioned_file2_uniçode'],
                log_environ=True,
                timeout=1200,
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['svn', 'update', '--non-interactive', '--no-auth-cache']
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_full_clean(self):
        self.setup_step(svn.SVN(repourl='http://svn.local/app/trunk', mode='full', method='clean'))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['svn', 'info', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['svn', 'status', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout(self.svn_st_xml_empty)
            .exit(0),
            ExpectShell(
                workdir='wkdir', command=['svn', 'update', '--non-interactive', '--no-auth-cache']
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_full_clean_given_revision(self):
        self.setup_step(
            svn.SVN(repourl='http://svn.local/app/trunk', mode='full', method='clean'),
            {"revision": '100'},
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['svn', 'info', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['svn', 'status', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout(self.svn_st_xml_empty)
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'update',
                    '--revision',
                    '100',
                    '--non-interactive',
                    '--no-auth-cache',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_full_not_updatable(self):
        self.setup_step(svn.SVN(repourl='http://svn.local/app/trunk', mode='full', method='clean'))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'checkout',
                    'http://svn.local/app/trunk',
                    '.',
                    '--non-interactive',
                    '--no-auth-cache',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_full_not_updatable_given_revision(self):
        self.setup_step(
            svn.SVN(repourl='http://svn.local/app/trunk', mode='full', method='clean'),
            {"revision": '100'},
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'checkout',
                    'http://svn.local/app/trunk',
                    '.',
                    '--revision',
                    '100',
                    '--non-interactive',
                    '--no-auth-cache',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_full_clean_old_rmdir(self):
        self.setup_step(svn.SVN(repourl='http://svn.local/app/trunk', mode='full', method='clean'))
        self.patch_workerVersionIsOlderThan(True)
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['svn', 'info', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['svn', 'status', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout(self.svn_st_xml)
            .exit(0),
            ExpectRmdir(
                dir='wkdir/svn_external_path/unversioned_file1', log_environ=True, timeout=1200
            ).exit(0),
            ExpectRmdir(
                dir='wkdir/svn_external_path/unversioned_file2_uniçode',
                log_environ=True,
                timeout=1200,
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['svn', 'update', '--non-interactive', '--no-auth-cache']
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_full_clean_new_rmdir(self):
        self.setup_step(svn.SVN(repourl='http://svn.local/app/trunk', mode='full', method='clean'))

        self.patch_workerVersionIsOlderThan(False)
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['svn', 'info', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['svn', 'status', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout(self.svn_st_xml)
            .exit(0),
            ExpectRmdir(
                dir=[
                    'wkdir/svn_external_path/unversioned_file1',
                    'wkdir/svn_external_path/unversioned_file2_uniçode',
                ],
                log_environ=True,
                timeout=1200,
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['svn', 'update', '--non-interactive', '--no-auth-cache']
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_full_copy(self):
        self.setup_step(
            svn.SVN(
                repourl='http://svn.local/app/trunk', mode='full', method='copy', codebase='app'
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(0),
            ExpectStat(file='source/app/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='source/app',
                command=['svn', 'info', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='source/app',
                command=['svn', 'update', '--non-interactive', '--no-auth-cache'],
            ).exit(0),
            ExpectCpdir(fromdir='source/app', todir='wkdir', log_environ=True).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', {'app': '100'}, 'SVN')
        return self.run_step()

    def test_mode_full_copy_given_revision(self):
        self.setup_step(
            svn.SVN(repourl='http://svn.local/app/trunk', mode='full', method='copy'),
            {"revision": '100'},
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(0),
            ExpectStat(file='source/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='source',
                command=['svn', 'info', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='source',
                command=[
                    'svn',
                    'update',
                    '--revision',
                    '100',
                    '--non-interactive',
                    '--no-auth-cache',
                ],
            ).exit(0),
            ExpectCpdir(fromdir='source', todir='wkdir', log_environ=True).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_full_export(self):
        self.setup_step(svn.SVN(repourl='http://svn.local/app/trunk', mode='full', method='export'))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(0),
            ExpectStat(file='source/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='source',
                command=['svn', 'info', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='source', command=['svn', 'update', '--non-interactive', '--no-auth-cache']
            ).exit(0),
            ExpectShell(workdir='', command=['svn', 'export', 'source', 'wkdir']).exit(0),
            ExpectShell(workdir='source', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_full_export_patch(self):
        self.setup_step(
            svn.SVN(repourl='http://svn.local/app/trunk', mode='full', method='export'),
            patch=(1, 'patch'),
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['svn', 'status', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout(self.svn_st_xml)
            .exit(0),
            ExpectRmdir(
                dir=[
                    'wkdir/svn_external_path/unversioned_file1',
                    'wkdir/svn_external_path/unversioned_file2_uniçode',
                ],
                log_environ=True,
                timeout=1200,
            ).exit(0),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(0),
            ExpectStat(file='source/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='source',
                command=['svn', 'info', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='source', command=['svn', 'update', '--non-interactive', '--no-auth-cache']
            ).exit(0),
            ExpectShell(workdir='', command=['svn', 'export', 'source', 'wkdir']).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                workerdest='.buildbot-diff',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                workerdest='.buildbot-patched',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'patch',
                    '-p1',
                    '--remove-empty-files',
                    '--force',
                    '--forward',
                    '-i',
                    '.buildbot-diff',
                ],
            ).exit(0),
            ExpectRmdir(dir='wkdir/.buildbot-diff', log_environ=True).exit(0),
            ExpectShell(workdir='source', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_full_export_patch_worker_2_16(self):
        self.setup_build(worker_version={'*': '2.16'})
        self.setup_step(
            svn.SVN(repourl='http://svn.local/app/trunk', mode='full', method='export'),
            patch=(1, 'patch'),
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['svn', 'status', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout(self.svn_st_xml)
            .exit(0),
            ExpectRmdir(
                dir=[
                    'wkdir/svn_external_path/unversioned_file1',
                    'wkdir/svn_external_path/unversioned_file2_uniçode',
                ],
                log_environ=True,
                timeout=1200,
            ).exit(0),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(0),
            ExpectStat(file='source/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='source',
                command=['svn', 'info', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='source', command=['svn', 'update', '--non-interactive', '--no-auth-cache']
            ).exit(0),
            ExpectShell(workdir='', command=['svn', 'export', 'source', 'wkdir']).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                slavedest='.buildbot-diff',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                slavedest='.buildbot-patched',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'patch',
                    '-p1',
                    '--remove-empty-files',
                    '--force',
                    '--forward',
                    '-i',
                    '.buildbot-diff',
                ],
            ).exit(0),
            ExpectRmdir(dir='wkdir/.buildbot-diff', log_environ=True).exit(0),
            ExpectShell(workdir='source', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_full_export_timeout(self):
        self.setup_step(
            svn.SVN(repourl='http://svn.local/app/trunk', timeout=1, mode='full', method='export')
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', timeout=1, command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1).exit(0),
            ExpectStat(file='source/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='source',
                timeout=1,
                command=['svn', 'info', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='source',
                timeout=1,
                command=['svn', 'update', '--non-interactive', '--no-auth-cache'],
            ).exit(0),
            ExpectShell(workdir='', timeout=1, command=['svn', 'export', 'source', 'wkdir']).exit(
                0
            ),
            ExpectShell(workdir='source', timeout=1, command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_full_export_given_revision(self):
        self.setup_step(
            svn.SVN(repourl='http://svn.local/app/trunk', mode='full', method='export'),
            {"revision": '100'},
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(0),
            ExpectStat(file='source/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='source',
                command=['svn', 'info', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='source',
                command=[
                    'svn',
                    'update',
                    '--revision',
                    '100',
                    '--non-interactive',
                    '--no-auth-cache',
                ],
            ).exit(0),
            ExpectShell(
                workdir='', command=['svn', 'export', '--revision', '100', 'source', 'wkdir']
            ).exit(0),
            ExpectShell(workdir='source', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_full_export_auth(self):
        self.setup_step(
            svn.SVN(
                repourl='http://svn.local/app/trunk',
                mode='full',
                method='export',
                username='svn_username',
                password='svn_password',
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(0),
            ExpectStat(file='source/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='source',
                command=[
                    'svn',
                    'info',
                    '--xml',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--username',
                    'svn_username',
                    '--password',
                    ('obfuscated', 'svn_password', 'XXXXXX'),
                ],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='source',
                command=[
                    'svn',
                    'update',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--username',
                    'svn_username',
                    '--password',
                    ('obfuscated', 'svn_password', 'XXXXXX'),
                ],
            ).exit(0),
            ExpectShell(
                workdir='',
                command=[
                    'svn',
                    'export',
                    '--username',
                    'svn_username',
                    '--password',
                    ('obfuscated', 'svn_password', 'XXXXXX'),
                    'source',
                    'wkdir',
                ],
            ).exit(0),
            ExpectShell(workdir='source', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_incremental_with_env(self):
        self.setup_step(
            svn.SVN(
                repourl='http://svn.local/app/trunk',
                mode='incremental',
                username='user',
                password='pass',
                extra_args=['--random'],
                env={'abc': '123'},
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version'], env={'abc': '123'}).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'info',
                    '--xml',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--username',
                    'user',
                    '--password',
                    ('obfuscated', 'pass', 'XXXXXX'),
                    '--random',
                ],
                env={'abc': '123'},
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'update',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--username',
                    'user',
                    '--password',
                    ('obfuscated', 'pass', 'XXXXXX'),
                    '--random',
                ],
                env={'abc': '123'},
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'], env={'abc': '123'})
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_mode_incremental_log_environ(self):
        self.setup_step(
            svn.SVN(
                repourl='http://svn.local/app/trunk',
                mode='incremental',
                username='user',
                password='pass',
                extra_args=['--random'],
                logEnviron=False,
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version'], log_environ=False).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=False).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=False).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'info',
                    '--xml',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--username',
                    'user',
                    '--password',
                    ('obfuscated', 'pass', 'XXXXXX'),
                    '--random',
                ],
                log_environ=False,
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'update',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--username',
                    'user',
                    '--password',
                    ('obfuscated', 'pass', 'XXXXXX'),
                    '--random',
                ],
                log_environ=False,
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'], log_environ=False)
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '100', 'SVN')
        return self.run_step()

    def test_command_fails(self):
        self.setup_step(
            svn.SVN(
                repourl='http://svn.local/app/trunk',
                mode='incremental',
                username='user',
                password='pass',
                extra_args=['--random'],
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'info',
                    '--xml',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--username',
                    'user',
                    '--password',
                    ('obfuscated', 'pass', 'XXXXXX'),
                    '--random',
                ],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'update',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--username',
                    'user',
                    '--password',
                    ('obfuscated', 'pass', 'XXXXXX'),
                    '--random',
                ],
            ).exit(1),
        )
        self.expect_outcome(result=FAILURE)
        return self.run_step()

    def test_bogus_svnversion(self):
        self.setup_step(
            svn.SVN(
                repourl='http://svn.local/app/trunk',
                mode='incremental',
                username='user',
                password='pass',
                extra_args=['--random'],
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'info',
                    '--xml',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--username',
                    'user',
                    '--password',
                    ('obfuscated', 'pass', 'XXXXXX'),
                    '--random',
                ],
            )
            .stdout(
                '<?xml version="1.0"?>'
                '<entry kind="dir" path="/a/b/c" revision="1">'
                '<url>http://svn.local/app/trunk</url>'
                '</entry>'
            )
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'update',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--username',
                    'user',
                    '--password',
                    ('obfuscated', 'pass', 'XXXXXX'),
                    '--random',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml']).stdout('1x0y0').exit(0),
        )
        self.expect_outcome(result=FAILURE)
        return self.run_step()

    def test_rmdir_fails_clobber(self):
        self.setup_step(
            svn.SVN(repourl='http://svn.local/app/trunk', mode='full', method='clobber')
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(1),
        )
        self.expect_outcome(result=FAILURE)
        return self.run_step()

    def test_rmdir_fails_copy(self):
        self.setup_step(svn.SVN(repourl='http://svn.local/app/trunk', mode='full', method='copy'))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(1),
        )
        self.expect_outcome(result=FAILURE)
        return self.run_step()

    def test_cpdir_fails_copy(self):
        self.setup_step(svn.SVN(repourl='http://svn.local/app/trunk', mode='full', method='copy'))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(0),
            ExpectStat(file='source/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='source',
                command=['svn', 'info', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='source', command=['svn', 'update', '--non-interactive', '--no-auth-cache']
            ).exit(0),
            ExpectCpdir(fromdir='source', todir='wkdir', log_environ=True).exit(1),
        )
        self.expect_outcome(result=FAILURE)
        return self.run_step()

    def test_rmdir_fails_purge(self):
        self.setup_step(
            svn.SVN(
                repourl='http://svn.local/app/trunk',
                mode='full',
                keep_on_purge=['svn_external_path/unversioned_file1'],
            )
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['svn', 'info', '--xml', '--non-interactive', '--no-auth-cache'],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'status',
                    '--xml',
                    '--no-ignore',
                    '--non-interactive',
                    '--no-auth-cache',
                ],
            )
            .stdout(self.svn_st_xml)
            .exit(0),
            ExpectRmdir(
                dir=['wkdir/svn_external_path/unversioned_file2_uniçode'],
                log_environ=True,
                timeout=1200,
            ).exit(1),
        )
        self.expect_outcome(result=FAILURE)
        return self.run_step()

    def test_worker_connection_lost(self):
        self.setup_step(
            svn.SVN(
                repourl='http://svn.local/app/trunk',
                mode='incremental',
                username='user',
                password='pass',
                extra_args=['--random'],
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).error(error.ConnectionLost())
        )
        self.expect_outcome(result=RETRY, state_string="update (retry)")
        return self.run_step()

    def test_empty_password(self):
        self.setup_step(
            svn.SVN(
                repourl='http://svn.local/app/trunk',
                mode='incremental',
                username='user',
                password='',
                extra_args=['--random'],
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'info',
                    '--xml',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--username',
                    'user',
                    '--password',
                    ('obfuscated', '', 'XXXXXX'),
                    '--random',
                ],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'update',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--username',
                    'user',
                    '--password',
                    ('obfuscated', '', 'XXXXXX'),
                    '--random',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_omit_password(self):
        self.setup_step(
            svn.SVN(
                repourl='http://svn.local/app/trunk',
                mode='incremental',
                username='user',
                extra_args=['--random'],
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['svn', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.svn', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'info',
                    '--xml',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--username',
                    'user',
                    '--random',
                ],
            )
            .stdout('<?xml version="1.0"?><url>http://svn.local/app/trunk</url>')
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'svn',
                    'update',
                    '--non-interactive',
                    '--no-auth-cache',
                    '--username',
                    'user',
                    '--random',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['svn', 'info', '--xml'])
            .stdout(self.svn_info_stdout_xml)
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()


class TestGetUnversionedFiles(unittest.TestCase):
    def test_getUnversionedFiles_does_not_list_externals(self):
        svn_st_xml = """<?xml version="1.0"?>
        <status>
            <target path=".">
                <entry path="svn_external_path">
                    <wc-status props="none" item="external">
                    </wc-status>
                </entry>
                <entry path="svn_external_path/unversioned_file">
                    <wc-status props="none" item="unversioned">
                    </wc-status>
                </entry>
            </target>
        </status>
        """
        unversioned_files = list(svn.SVN.getUnversionedFiles(svn_st_xml, []))
        self.assertEqual(["svn_external_path/unversioned_file"], unversioned_files)

    def test_getUnversionedFiles_does_not_list_missing(self):
        svn_st_xml = """<?xml version="1.0"?>
        <status>
            <target path=".">
                <entry path="missing_file">
                    <wc-status props="none" item="missing"></wc-status>
                </entry>
            </target>
        </status>
        """
        unversioned_files = list(svn.SVN.getUnversionedFiles(svn_st_xml, []))
        self.assertEqual([], unversioned_files)

    def test_getUnversionedFiles_corrupted_xml(self):
        svn_st_xml_corrupt = """<?xml version="1.0"?>
            <target path=".">
                <entry path="svn_external_path">
                    <wc-status props="none" item="external">
                    </wc-status>
                </entry>
                <entry path="svn_external_path/unversioned_file">
                    <wc-status props="none" item="unversioned">
                    </wc-status>
                </entry>
            </target>
        </status>
        """
        with self.assertRaises(buildstep.BuildStepFailed):
            list(svn.SVN.getUnversionedFiles(svn_st_xml_corrupt, []))

    def test_getUnversionedFiles_no_path(self):
        svn_st_xml = """<?xml version="1.0"?>
        <status>
            <target path=".">
                <entry path="svn_external_path">
                    <wc-status props="none" item="external">
                    </wc-status>
                </entry>
                <entry>
                    <wc-status props="none" item="unversioned">
                    </wc-status>
                </entry>
            </target>
        </status>
        """
        unversioned_files = list(svn.SVN.getUnversionedFiles(svn_st_xml, []))
        self.assertEqual([], unversioned_files)

    def test_getUnversionedFiles_no_item(self):
        svn_st_xml = """<?xml version="1.0"?>
        <status>
            <target path=".">
                <entry path="svn_external_path">
                    <wc-status props="none" item="external">
                    </wc-status>
                </entry>
                <entry path="svn_external_path/unversioned_file">
                    <wc-status props="none">
                    </wc-status>
                </entry>
            </target>
        </status>
        """
        unversioned_files = list(svn.SVN.getUnversionedFiles(svn_st_xml, []))
        self.assertEqual(["svn_external_path/unversioned_file"], unversioned_files)

    def test_getUnversionedFiles_unicode(self):
        svn_st_xml = """<?xml version="1.0"?>
        <status>
            <target path=".">
                <entry
                   path="Path/To/Content/Developers/François">
                    <wc-status
                       item="unversioned"
                       props="none">
                    </wc-status>
                </entry>
            </target>
        </status>
        """
        unversioned_files = list(svn.SVN.getUnversionedFiles(svn_st_xml, []))
        self.assertEqual(["Path/To/Content/Developers/François"], unversioned_files)


class TestSvnUriCanonicalize(unittest.TestCase):
    @parameterized.expand([
        ("empty", "", ""),
        ("canonical", "http://foo.com/bar", "http://foo.com/bar"),
        ("lc_scheme", "hTtP://foo.com/bar", "http://foo.com/bar"),
        ("trailing_dot", "http://foo.com./bar", "http://foo.com/bar"),
        ("lc_hostname", "http://foO.COm/bar", "http://foo.com/bar"),
        ("lc_hostname_with_user", "http://Jimmy@fOO.Com/bar", "http://Jimmy@foo.com/bar"),
        (
            "lc_hostname_with_user_pass",
            "http://Jimmy:Sekrit@fOO.Com/bar",
            "http://Jimmy:Sekrit@foo.com/bar",
        ),
        ("trailing_slash", "http://foo.com/bar/", "http://foo.com/bar"),
        ("trailing_slash_scheme", "http://", "http://"),
        ("trailing_slash_hostname", "http://foo.com/", "http://foo.com"),
        ("trailing_double_slash", "http://foo.com/x//", "http://foo.com/x"),
        ("double_slash", "http://foo.com/x//y", "http://foo.com/x/y"),
        ("slash", "/", "/"),
        ("dot", "http://foo.com/x/./y", "http://foo.com/x/y"),
        ("dot_dot", "http://foo.com/x/../y", "http://foo.com/y"),
        ("double_dot_dot", "http://foo.com/x/y/../../z", "http://foo.com/z"),
        ("dot_dot_root", "http://foo.com/../x/y", "http://foo.com/x/y"),
        (
            "quote_spaces",
            "svn+ssh://user@host:123/My Stuff/file.doc",
            "svn+ssh://user@host:123/My%20Stuff/file.doc",
        ),
        ("remove_port_80", "http://foo.com:80/bar", "http://foo.com/bar"),
        ("dont_remove_port_80", "https://foo.com:80/bar", "https://foo.com:80/bar"),  # not http
        ("remove_port_443", "https://foo.com:443/bar", "https://foo.com/bar"),
        ("dont_remove_port_443", "svn://foo.com:443/bar", "svn://foo.com:443/bar"),  # not https
        ("remove_port_3690", "svn://foo.com:3690/bar", "svn://foo.com/bar"),
        ("dont_remove_port_3690", "http://foo.com:3690/bar", "http://foo.com:3690/bar"),  # not svn
        ("dont_remove_port_other", "https://foo.com:2093/bar", "https://foo.com:2093/bar"),
        ("quote_funny_chars", "http://foo.com/\x10\xe6%", "http://foo.com/%10%E6%25"),
        (
            "overquoted",
            "http://foo.com/%68%65%6c%6c%6f%20%77%6f%72%6c%64",
            "http://foo.com/hello%20world",
        ),
    ])
    def test_svn_uri(self, name, input, exp):
        self.assertEqual(svn.SVN.svnUriCanonicalize(input), exp)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING

from twisted.trial import unittest

from buildbot.db import compression
from buildbot.db.logs import RawCompressor

if TYPE_CHECKING:
    from typing import ClassVar

    from buildbot.db.compression.protocol import CompressorInterface


class TestRawCompressor(unittest.TestCase):
    # default no-op compressor
    CompressorCls: ClassVar[type[CompressorInterface]] = RawCompressor

    def test_dumps_read(self) -> None:
        if not self.CompressorCls.available:
            raise unittest.SkipTest(f"Compressor '{self.CompressorCls.name}' is unavailable")

        data = b'xy' * 10000
        compressed_data = self.CompressorCls.dumps(data)
        self.assertEqual(data, self.CompressorCls.read(compressed_data))

    def test_compressobj_read(self) -> None:
        if not self.CompressorCls.available:
            raise unittest.SkipTest(f"Compressor '{self.CompressorCls.name}' is unavailable")

        input_buffer = [f'xy{idx}'.encode() * 10000 for idx in range(10)]

        compress_obj = self.CompressorCls.CompressObj()

        def _test() -> None:
            result_buffer = [compress_obj.compress(e) for e in input_buffer]
            result_buffer.append(compress_obj.flush())

            self.assertEqual(
                b''.join(input_buffer), self.CompressorCls.read(b''.join(result_buffer))
            )

        _test()

        # make sure re-using the same compress obj works
        _test()


class TestGZipCompressor(TestRawCompressor):
    CompressorCls = compression.GZipCompressor


class TestBZipCompressor(TestRawCompressor):
    CompressorCls = compression.BZipCompressor


class TestLZ4Compressor(TestRawCompressor):
    CompressorCls = compression.LZ4Compressor


class TestBrotliCompressor(TestRawCompressor):
    CompressorCls = compression.BrotliCompressor


class TestZStdCompressor(TestRawCompressor):
    CompressorCls = compression.ZStdCompressor



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.trial import unittest

from buildbot import config
from buildbot.test.fake import httpclientservice as fakehttpclientservice
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import www
from buildbot.www import auth
from buildbot.www import avatar


class TestAvatar(avatar.AvatarBase):
    def getUserAvatar(self, email, username, size, defaultAvatarUrl):
        user_avatar = f'{email!r} {size!r} {defaultAvatarUrl!r}'.encode()
        return defer.succeed((b"image/png", user_avatar))


class AvatarResource(TestReactorMixin, www.WwwTestMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()

    @defer.inlineCallbacks
    def test_default(self):
        master = yield self.make_master(url='http://a/b/', auth=auth.NoAuth(), avatar_methods=[])
        rsrc = avatar.AvatarResource(master)
        rsrc.reconfigResource(master.config)

        res = yield self.render_resource(rsrc, b'/')
        self.assertEqual(res, {"redirected": avatar.AvatarResource.defaultAvatarUrl})

    @defer.inlineCallbacks
    def test_gravatar(self):
        master = yield self.make_master(
            url='http://a/b/', auth=auth.NoAuth(), avatar_methods=[avatar.AvatarGravatar()]
        )
        rsrc = avatar.AvatarResource(master)
        rsrc.reconfigResource(master.config)

        res = yield self.render_resource(rsrc, b'/?email=foo')
        self.assertEqual(
            res,
            {
                "redirected": b'//www.gravatar.com/avatar/acbd18db4cc2f85ce'
                b'def654fccc4a4d8?d=retro&s=32'
            },
        )

    @defer.inlineCallbacks
    def test_avatar_call(self):
        master = yield self.make_master(
            url='http://a/b/', auth=auth.NoAuth(), avatar_methods=[TestAvatar()]
        )
        rsrc = avatar.AvatarResource(master)
        rsrc.reconfigResource(master.config)

        res = yield self.render_resource(rsrc, b'/?email=foo')
        self.assertEqual(res, b"b'foo' 32 'http://a/b/img/nobody.png'")

    @defer.inlineCallbacks
    def test_custom_size(self):
        master = yield self.make_master(
            url='http://a/b/', auth=auth.NoAuth(), avatar_methods=[TestAvatar()]
        )
        rsrc = avatar.AvatarResource(master)
        rsrc.reconfigResource(master.config)

        res = yield self.render_resource(rsrc, b'/?email=foo&size=64')
        self.assertEqual(res, b"b'foo' 64 'http://a/b/img/nobody.png'")

    @defer.inlineCallbacks
    def test_invalid_size(self):
        master = yield self.make_master(
            url='http://a/b/', auth=auth.NoAuth(), avatar_methods=[TestAvatar()]
        )
        rsrc = avatar.AvatarResource(master)
        rsrc.reconfigResource(master.config)

        res = yield self.render_resource(rsrc, b'/?email=foo&size=abcd')
        self.assertEqual(res, b"b'foo' 32 'http://a/b/img/nobody.png'")

    @defer.inlineCallbacks
    def test_custom_not_found(self):
        # use gravatar if the custom avatar fail to return a response
        class CustomAvatar(avatar.AvatarBase):
            def getUserAvatar(self, email, username, size, defaultAvatarUrl):
                return defer.succeed(None)

        master = yield self.make_master(
            url=b'http://a/b/',
            auth=auth.NoAuth(),
            avatar_methods=[CustomAvatar(), avatar.AvatarGravatar()],
        )
        rsrc = avatar.AvatarResource(master)
        rsrc.reconfigResource(master.config)

        res = yield self.render_resource(rsrc, b'/?email=foo')
        self.assertEqual(
            res,
            {
                "redirected": b'//www.gravatar.com/avatar/acbd18db4cc2f85ce'
                b'def654fccc4a4d8?d=retro&s=32'
            },
        )


github_username_search_reply = {
    "login": "defunkt",
    "id": 42424242,
    "node_id": "MDQ6VXNlcjQyNDI0MjQy",
    "avatar_url": "https://avatars3.githubusercontent.com/u/42424242?v=4",
    "gravatar_id": "",
    "url": "https://api.github.com/users/defunkt",
    "html_url": "https://github.com/defunkt",
    "followers_url": "https://api.github.com/users/defunkt/followers",
    "following_url": "https://api.github.com/users/defunkt/following{/other_user}",
    "gists_url": "https://api.github.com/users/defunkt/gists{/gist_id}",
    "starred_url": "https://api.github.com/users/defunkt/starred{/owner}{/repo}",
    "subscriptions_url": "https://api.github.com/users/defunkt/subscriptions",
    "organizations_url": "https://api.github.com/users/defunkt/orgs",
    "repos_url": "https://api.github.com/users/defunkt/repos",
    "events_url": "https://api.github.com/users/defunkt/events{/privacy}",
    "received_events_url": "https://api.github.com/users/defunkt/received_events",
    "type": "User",
    "site_admin": False,
    "name": "Defunkt User",
    "company": None,
    "blog": "",
    "location": None,
    "email": None,
    "hireable": None,
    "bio": None,
    "twitter_username": None,
    "public_repos": 1,
    "public_gists": 1,
    "followers": 1,
    "following": 1,
    "created_at": "2000-01-01T00:00:00Z",
    "updated_at": "2021-01-01T00:00:00Z",
}

github_username_not_found_reply = {
    "message": "Not Found",
    "documentation_url": "https://docs.github.com/rest/reference/users#get-a-user",
}

github_email_search_reply = {
    "total_count": 1,
    "incomplete_results": False,
    "items": [
        {
            "login": "defunkt",
            "id": 42424242,
            "node_id": "MDQ6VXNlcjQyNDI0MjQy",
            "avatar_url": "https://avatars3.githubusercontent.com/u/42424242?v=4",
            "gravatar_id": "",
            "url": "https://api.github.com/users/defunkt",
            "html_url": "https://github.com/defunkt",
            "followers_url": "https://api.github.com/users/defunkt/followers",
            "following_url": "https://api.github.com/users/defunkt/following{/other_user}",
            "gists_url": "https://api.github.com/users/defunkt/gists{/gist_id}",
            "starred_url": "https://api.github.com/users/defunkt/starred{/owner}{/repo}",
            "subscriptions_url": "https://api.github.com/users/defunkt/subscriptions",
            "organizations_url": "https://api.github.com/users/defunkt/orgs",
            "repos_url": "https://api.github.com/users/defunkt/repos",
            "events_url": "https://api.github.com/users/defunkt/events{/privacy}",
            "received_events_url": "https://api.github.com/users/defunkt/received_events",
            "type": "User",
            "site_admin": False,
            "score": 1.0,
        }
    ],
}

github_email_search_not_found_reply = {"total_count": 0, "incomplete_results": False, "items": []}

github_commit_search_reply = {
    "total_count": 1,
    "incomplete_results": False,
    "items": [
        {
            "url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
            "commits/1111111111111111111111111111111111111111",
            "sha": "1111111111111111111111111111111111111111",
            "node_id": "MDY6Q29tbWl0NDM0MzQzNDM6MTExMTExMTExMTExMTExMTExMTExMTExMTExMTExMTExMTExMTEx",
            "html_url": "https://github.com/defunkt-org/defunkt-repo/"
            "commit/1111111111111111111111111111111111111111",
            "comments_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
            "commits/1111111111111111111111111111111111111111/comments",
            "commit": {
                "url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "git/commits/1111111111111111111111111111111111111111",
                "author": {
                    "date": "2021-01-01T01:01:01.000-01:00",
                    "name": "Defunkt User",
                    "email": "defunkt@defunkt.com",
                },
                "committer": {
                    "date": "2021-01-01T01:01:01.000-01:00",
                    "name": "Defunkt User",
                    "email": "defunkt@defunkt.com",
                },
                "message": "defunkt message",
                "tree": {
                    "url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                    "git/trees/2222222222222222222222222222222222222222",
                    "sha": "2222222222222222222222222222222222222222",
                },
                "comment_count": 0,
            },
            "author": {
                "login": "defunkt",
                "id": 42424242,
                "node_id": "MDQ6VXNlcjQyNDI0MjQy",
                "avatar_url": "https://avatars3.githubusercontent.com/u/42424242?v=4",
                "gravatar_id": "",
                "url": "https://api.github.com/users/defunkt",
                "html_url": "https://github.com/defunkt",
                "followers_url": "https://api.github.com/users/defunkt/followers",
                "following_url": "https://api.github.com/users/defunkt/following{/other_user}",
                "gists_url": "https://api.github.com/users/defunkt/gists{/gist_id}",
                "starred_url": "https://api.github.com/users/defunkt/starred{/owner}{/repo}",
                "subscriptions_url": "https://api.github.com/users/defunkt/subscriptions",
                "organizations_url": "https://api.github.com/users/defunkt/orgs",
                "repos_url": "https://api.github.com/users/defunkt/repos",
                "events_url": "https://api.github.com/users/defunkt/events{/privacy}",
                "received_events_url": "https://api.github.com/users/defunkt/received_events",
                "type": "User",
                "site_admin": False,
            },
            "committer": {
                "login": "defunkt",
                "id": 42424242,
                "node_id": "MDQ6VXNlcjQyNDI0MjQy",
                "avatar_url": "https://avatars3.githubusercontent.com/u/42424242?v=4",
                "gravatar_id": "",
                "url": "https://api.github.com/users/defunkt",
                "html_url": "https://github.com/defunkt",
                "followers_url": "https://api.github.com/users/defunkt/followers",
                "following_url": "https://api.github.com/users/defunkt/following{/other_user}",
                "gists_url": "https://api.github.com/users/defunkt/gists{/gist_id}",
                "starred_url": "https://api.github.com/users/defunkt/starred{/owner}{/repo}",
                "subscriptions_url": "https://api.github.com/users/defunkt/subscriptions",
                "organizations_url": "https://api.github.com/users/defunkt/orgs",
                "repos_url": "https://api.github.com/users/defunkt/repos",
                "events_url": "https://api.github.com/users/defunkt/events{/privacy}",
                "received_events_url": "https://api.github.com/users/defunkt/received_events",
                "type": "User",
                "site_admin": False,
            },
            "parents": [
                {
                    "url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                    "commits/3333333333333333333333333333333333333333",
                    "html_url": "https://github.com/defunkt-org/defunkt-repo/"
                    "commit/3333333333333333333333333333333333333333",
                    "sha": "3333333333333333333333333333333333333333",
                }
            ],
            "repository": {
                "id": 43434343,
                "node_id": "MDEwOlJlcG9zaXRvcnk0MzQzNDM0Mw==",
                "name": "defunkt-repo",
                "full_name": "defunkt-org/defunkt-repo",
                "private": False,
                "owner": {
                    "login": "defunkt-org",
                    "id": 44444444,
                    "node_id": "MDEyOk9yZ2FuaXphdGlvbjQ0NDQ0NDQ0",
                    "avatar_url": "https://avatars2.githubusercontent.com/u/44444444?v=4",
                    "gravatar_id": "",
                    "url": "https://api.github.com/users/defunkt-org",
                    "html_url": "https://github.com/defunkt-org",
                    "followers_url": "https://api.github.com/users/defunkt-org/followers",
                    "following_url": "https://api.github.com/users/defunkt-org/"
                    "following{/other_user}",
                    "gists_url": "https://api.github.com/users/defunkt-org/gists{/gist_id}",
                    "starred_url": "https://api.github.com/users/defunkt-org/"
                    "starred{/owner}{/repo}",
                    "subscriptions_url": "https://api.github.com/users/defunkt-org/subscriptions",
                    "organizations_url": "https://api.github.com/users/defunkt-org/orgs",
                    "repos_url": "https://api.github.com/users/defunkt-org/repos",
                    "events_url": "https://api.github.com/users/defunkt-org/events{/privacy}",
                    "received_events_url": "https://api.github.com/users/defunkt-org/"
                    "received_events",
                    "type": "Organization",
                    "site_admin": False,
                },
                "html_url": "https://github.com/defunkt-org/defunkt-repo",
                "description": "defunkt project",
                "fork": False,
                "url": "https://api.github.com/repos/defunkt-org/defunkt-repo",
                "forks_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/forks",
                "keys_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/keys{/key_id}",
                "collaborators_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "collaborators{/collaborator}",
                "teams_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/teams",
                "hooks_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/hooks",
                "issue_events_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "issues/events{/number}",
                "events_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/events",
                "assignees_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "assignees{/user}",
                "branches_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "branches{/branch}",
                "tags_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/tags",
                "blobs_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "git/blobs{/sha}",
                "git_tags_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "git/tags{/sha}",
                "git_refs_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "git/refs{/sha}",
                "trees_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "git/trees{/sha}",
                "statuses_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "statuses/{sha}",
                "languages_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/languages",
                "stargazers_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "stargazers",
                "contributors_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "contributors",
                "subscribers_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "subscribers",
                "subscription_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "subscription",
                "commits_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "commits{/sha}",
                "git_commits_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "git/commits{/sha}",
                "comments_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "comments{/number}",
                "issue_comment_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "issues/comments{/number}",
                "contents_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "contents/{+path}",
                "compare_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "compare/{base}...{head}",
                "merges_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/merges",
                "archive_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "{archive_format}{/ref}",
                "downloads_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/downloads",
                "issues_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "issues{/number}",
                "pulls_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/pulls{/number}",
                "milestones_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "milestones{/number}",
                "notifications_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "notifications{?since,all,participating}",
                "labels_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/labels{/name}",
                "releases_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "releases{/id}",
                "deployments_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "deployments",
            },
            "score": 1.0,
        }
    ],
}

github_commit_search_no_user_reply = {
    "total_count": 1,
    "incomplete_results": False,
    "items": [
        {
            "url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
            "commits/1111111111111111111111111111111111111111",
            "sha": "1111111111111111111111111111111111111111",
            "node_id": "MDY6Q29tbWl0NDM0MzQzNDM6MTExMTExMTExMTExMTExMTExMTExMTExMTExMTExMTExMTExMTEx",
            "html_url": "https://github.com/defunkt-org/defunkt-repo/"
            "commit/1111111111111111111111111111111111111111",
            "comments_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
            "commits/1111111111111111111111111111111111111111/comments",
            "commit": {
                "url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "git/commits/1111111111111111111111111111111111111111",
                "author": {
                    "date": "2021-01-01T01:01:01.000-01:00",
                    "name": "Defunkt User",
                    "email": "defunkt@defunkt.com",
                },
                "committer": {
                    "date": "2021-01-01T01:01:01.000-01:00",
                    "name": "Defunkt User",
                    "email": "defunkt@defunkt.com",
                },
                "message": "defunkt message",
                "tree": {
                    "url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                    "git/trees/2222222222222222222222222222222222222222",
                    "sha": "2222222222222222222222222222222222222222",
                },
                "comment_count": 0,
            },
            "author": None,
            "committer": None,
            "parents": [
                {
                    "url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                    "commits/3333333333333333333333333333333333333333",
                    "html_url": "https://github.com/defunkt-org/defunkt-repo/"
                    "commit/3333333333333333333333333333333333333333",
                    "sha": "3333333333333333333333333333333333333333",
                }
            ],
            "repository": {
                "id": 43434343,
                "node_id": "MDEwOlJlcG9zaXRvcnk0MzQzNDM0Mw==",
                "name": "defunkt-repo",
                "full_name": "defunkt-org/defunkt-repo",
                "private": False,
                "owner": {
                    "login": "defunkt-org",
                    "id": 44444444,
                    "node_id": "MDEyOk9yZ2FuaXphdGlvbjQ0NDQ0NDQ0",
                    "avatar_url": "https://avatars2.githubusercontent.com/u/44444444?v=4",
                    "gravatar_id": "",
                    "url": "https://api.github.com/users/defunkt-org",
                    "html_url": "https://github.com/defunkt-org",
                    "followers_url": "https://api.github.com/users/defunkt-org/followers",
                    "following_url": "https://api.github.com/users/defunkt-org/"
                    "following{/other_user}",
                    "gists_url": "https://api.github.com/users/defunkt-org/gists{/gist_id}",
                    "starred_url": "https://api.github.com/users/defunkt-org/"
                    "starred{/owner}{/repo}",
                    "subscriptions_url": "https://api.github.com/users/defunkt-org/subscriptions",
                    "organizations_url": "https://api.github.com/users/defunkt-org/orgs",
                    "repos_url": "https://api.github.com/users/defunkt-org/repos",
                    "events_url": "https://api.github.com/users/defunkt-org/events{/privacy}",
                    "received_events_url": "https://api.github.com/users/defunkt-org/"
                    "received_events",
                    "type": "Organization",
                    "site_admin": False,
                },
                "html_url": "https://github.com/defunkt-org/defunkt-repo",
                "description": "defunkt project",
                "fork": False,
                "url": "https://api.github.com/repos/defunkt-org/defunkt-repo",
                "forks_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/forks",
                "keys_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/keys{/key_id}",
                "collaborators_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "collaborators{/collaborator}",
                "teams_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/teams",
                "hooks_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/hooks",
                "issue_events_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "issues/events{/number}",
                "events_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/events",
                "assignees_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "assignees{/user}",
                "branches_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "branches{/branch}",
                "tags_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/tags",
                "blobs_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "git/blobs{/sha}",
                "git_tags_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "git/tags{/sha}",
                "git_refs_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "git/refs{/sha}",
                "trees_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "git/trees{/sha}",
                "statuses_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "statuses/{sha}",
                "languages_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/languages",
                "stargazers_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "stargazers",
                "contributors_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "contributors",
                "subscribers_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "subscribers",
                "subscription_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "subscription",
                "commits_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "commits{/sha}",
                "git_commits_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "git/commits{/sha}",
                "comments_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "comments{/number}",
                "issue_comment_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "issues/comments{/number}",
                "contents_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "contents/{+path}",
                "compare_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "compare/{base}...{head}",
                "merges_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/merges",
                "archive_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "{archive_format}{/ref}",
                "downloads_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/downloads",
                "issues_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "issues{/number}",
                "pulls_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/pulls{/number}",
                "milestones_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "milestones{/number}",
                "notifications_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "notifications{?since,all,participating}",
                "labels_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/labels{/name}",
                "releases_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "releases{/id}",
                "deployments_url": "https://api.github.com/repos/defunkt-org/defunkt-repo/"
                "deployments",
            },
            "score": 1.0,
        }
    ],
}

github_commit_search_not_found_reply = {"total_count": 0, "incomplete_results": False, "items": []}


class GitHubAvatar(TestReactorMixin, www.WwwTestMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()

        master = yield self.make_master(
            url='http://a/b/',
            auth=auth.NoAuth(),
            avatar_methods=[avatar.AvatarGitHub(token="abcd")],
        )

        self.rsrc = avatar.AvatarResource(master)
        self.rsrc.reconfigResource(master.config)

        headers = {
            'User-Agent': 'Buildbot',
            'Authorization': 'token abcd',
        }
        self._http = yield fakehttpclientservice.HTTPClientService.getService(
            master,
            self,
            avatar.AvatarGitHub.DEFAULT_GITHUB_API_URL,
            headers=headers,
            debug=False,
            verify=True,
        )
        yield self.master.startService()
        self.addCleanup(self.master.stopService)

    @defer.inlineCallbacks
    def test_username(self):
        username_search_endpoint = '/users/defunkt'
        self._http.expect(
            'get',
            username_search_endpoint,
            content_json=github_username_search_reply,
            headers={'Accept': 'application/vnd.github.v3+json'},
        )
        res = yield self.render_resource(self.rsrc, b'/?username=defunkt')
        self.assertEqual(
            res, {"redirected": b'https://avatars3.githubusercontent.com/u/42424242?v=4&s=32'}
        )

    @defer.inlineCallbacks
    def test_username_not_found(self):
        username_search_endpoint = '/users/inexistent'
        self._http.expect(
            'get',
            username_search_endpoint,
            code=404,
            content_json=github_username_not_found_reply,
            headers={'Accept': 'application/vnd.github.v3+json'},
        )
        res = yield self.render_resource(self.rsrc, b'/?username=inexistent')
        self.assertEqual(res, {"redirected": b'img/nobody.png'})

    @defer.inlineCallbacks
    def test_username_error(self):
        username_search_endpoint = '/users/error'
        self._http.expect(
            'get',
            username_search_endpoint,
            code=500,
            headers={'Accept': 'application/vnd.github.v3+json'},
        )
        res = yield self.render_resource(self.rsrc, b'/?username=error')
        self.assertEqual(res, {"redirected": b'img/nobody.png'})

    @defer.inlineCallbacks
    def test_username_cached(self):
        username_search_endpoint = '/users/defunkt'
        self._http.expect(
            'get',
            username_search_endpoint,
            content_json=github_username_search_reply,
            headers={'Accept': 'application/vnd.github.v3+json'},
        )
        res = yield self.render_resource(self.rsrc, b'/?username=defunkt')
        self.assertEqual(
            res, {"redirected": b'https://avatars3.githubusercontent.com/u/42424242?v=4&s=32'}
        )
        # Second request will give same result but without an HTTP request
        res = yield self.render_resource(self.rsrc, b'/?username=defunkt')
        self.assertEqual(
            res, {"redirected": b'https://avatars3.githubusercontent.com/u/42424242?v=4&s=32'}
        )

    @defer.inlineCallbacks
    def test_email(self):
        email_search_endpoint = '/search/users?q=defunkt%40defunkt.com+in%3Aemail'
        self._http.expect(
            'get',
            email_search_endpoint,
            content_json=github_email_search_reply,
            headers={'Accept': 'application/vnd.github.v3+json'},
        )
        res = yield self.render_resource(self.rsrc, b'/?email=defunkt@defunkt.com')
        self.assertEqual(
            res, {"redirected": b'https://avatars3.githubusercontent.com/u/42424242?v=4&s=32'}
        )

    @defer.inlineCallbacks
    def test_email_commit(self):
        email_search_endpoint = '/search/users?q=defunkt%40defunkt.com+in%3Aemail'
        self._http.expect(
            'get',
            email_search_endpoint,
            content_json=github_email_search_not_found_reply,
            headers={'Accept': 'application/vnd.github.v3+json'},
        )
        commit_search_endpoint = (
            '/search/commits?per_page=1&q=author-email%3Adefunkt%40defunkt.com&sort=committer-date'
        )
        self._http.expect(
            'get',
            commit_search_endpoint,
            content_json=github_commit_search_reply,
            headers={
                'Accept': 'application/vnd.github.v3+json,application/vnd.github.cloak-preview'
            },
        )
        res = yield self.render_resource(self.rsrc, b'/?email=defunkt@defunkt.com')
        self.assertEqual(
            res, {"redirected": b'https://avatars3.githubusercontent.com/u/42424242?v=4&s=32'}
        )

    @defer.inlineCallbacks
    def test_email_commit_no_user(self):
        email_search_endpoint = '/search/users?q=defunkt%40defunkt.com+in%3Aemail'
        self._http.expect(
            'get',
            email_search_endpoint,
            content_json=github_email_search_not_found_reply,
            headers={'Accept': 'application/vnd.github.v3+json'},
        )
        commit_search_endpoint = (
            '/search/commits?per_page=1&q=author-email%3Adefunkt%40defunkt.com&sort=committer-date'
        )
        self._http.expect(
            'get',
            commit_search_endpoint,
            content_json=github_commit_search_no_user_reply,
            headers={
                'Accept': 'application/vnd.github.v3+json,application/vnd.github.cloak-preview'
            },
        )
        res = yield self.render_resource(self.rsrc, b'/?email=defunkt@defunkt.com')
        self.assertEqual(res, {"redirected": b'img/nobody.png'})

    @defer.inlineCallbacks
    def test_email_not_found(self):
        email_search_endpoint = '/search/users?q=notfound%40defunkt.com+in%3Aemail'
        self._http.expect(
            'get',
            email_search_endpoint,
            content_json=github_email_search_not_found_reply,
            headers={'Accept': 'application/vnd.github.v3+json'},
        )
        commit_search_endpoint = (
            '/search/commits?per_page=1&q=author-email%3Anotfound%40defunkt.com&sort=committer-date'
        )
        self._http.expect(
            'get',
            commit_search_endpoint,
            content_json=github_commit_search_not_found_reply,
            headers={
                'Accept': 'application/vnd.github.v3+json,application/vnd.github.cloak-preview'
            },
        )
        res = yield self.render_resource(self.rsrc, b'/?email=notfound@defunkt.com')
        self.assertEqual(res, {"redirected": b'img/nobody.png'})

    @defer.inlineCallbacks
    def test_email_error(self):
        email_search_endpoint = '/search/users?q=error%40defunkt.com+in%3Aemail'
        self._http.expect(
            'get',
            email_search_endpoint,
            code=500,
            headers={'Accept': 'application/vnd.github.v3+json'},
        )
        commit_search_endpoint = (
            '/search/commits?per_page=1&q=author-email%3Aerror%40defunkt.com&sort=committer-date'
        )
        self._http.expect(
            'get',
            commit_search_endpoint,
            code=500,
            headers={
                'Accept': 'application/vnd.github.v3+json,application/vnd.github.cloak-preview'
            },
        )
        res = yield self.render_resource(self.rsrc, b'/?email=error@defunkt.com')
        self.assertEqual(res, {"redirected": b'img/nobody.png'})


class GitHubAvatarBasicAuth(TestReactorMixin, www.WwwTestMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()

        avatar_method = avatar.AvatarGitHub(client_id="oauth_id", client_secret="oauth_secret")
        master = yield self.make_master(
            url='http://a/b/', auth=auth.NoAuth(), avatar_methods=[avatar_method]
        )

        self.rsrc = avatar.AvatarResource(master)
        self.rsrc.reconfigResource(master.config)

        headers = {
            'User-Agent': 'Buildbot',
            # oauth_id:oauth_secret in Base64
            'Authorization': 'basic b2F1dGhfaWQ6b2F1dGhfc2VjcmV0',
        }
        self._http = yield fakehttpclientservice.HTTPClientService.getService(
            master,
            self,
            avatar.AvatarGitHub.DEFAULT_GITHUB_API_URL,
            headers=headers,
            debug=False,
            verify=True,
        )
        yield self.master.startService()
        self.addCleanup(self.master.stopService)

    def test_incomplete_credentials(self):
        with self.assertRaises(config.ConfigErrors):
            avatar.AvatarGitHub(client_id="oauth_id")
        with self.assertRaises(config.ConfigErrors):
            avatar.AvatarGitHub(client_secret="oauth_secret")

    def test_token_and_client_credentials(self):
        with self.assertRaises(config.ConfigErrors):
            avatar.AvatarGitHub(client_id="oauth_id", client_secret="oauth_secret", token="token")

    @defer.inlineCallbacks
    def test_username(self):
        username_search_endpoint = '/users/defunkt'
        self._http.expect(
            'get',
            username_search_endpoint,
            content_json=github_username_search_reply,
            headers={'Accept': 'application/vnd.github.v3+json'},
        )
        res = yield self.render_resource(self.rsrc, b'/?username=defunkt')
        self.assertEqual(
            res, {'redirected': b'https://avatars3.githubusercontent.com/u/42424242?v=4&s=32'}
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import getpass
import os
import sys
from io import StringIO
from unittest import mock

from twisted.python import log
from twisted.python import runtime
from twisted.python import usage
from twisted.trial import unittest

from buildbot.scripts import base
from buildbot.scripts import runner
from buildbot.test.util import misc


class OptionsMixin:
    def setUpOptions(self):
        self.options_file = {}
        self.patch(base.SubcommandOptions, 'loadOptionsFile', lambda other_self: self.options_file)

    def assertOptions(self, opts, exp):
        got = {k: opts[k] for k in exp}
        if got != exp:
            msg = []
            for k in exp:
                if opts[k] != exp[k]:
                    msg.append(f" {k}: expected {exp[k]!r}, got {opts[k]!r}")
            self.fail("did not get expected options\n" + ("\n".join(msg)))


class TestUpgradeMasterOptions(OptionsMixin, unittest.TestCase):
    def setUp(self):
        self.setUpOptions()

    def parse(self, *args):
        self.opts = runner.UpgradeMasterOptions()
        self.opts.parseOptions(args)
        return self.opts

    def test_synopsis(self):
        opts = runner.UpgradeMasterOptions()
        self.assertIn('buildbot upgrade-master', opts.getSynopsis())

    def test_defaults(self):
        opts = self.parse()
        exp = {"quiet": False, "replace": False}
        self.assertOptions(opts, exp)

    def test_short(self):
        opts = self.parse('-q', '-r')
        exp = {"quiet": True, "replace": True}
        self.assertOptions(opts, exp)

    def test_long(self):
        opts = self.parse('--quiet', '--replace')
        exp = {"quiet": True, "replace": True}
        self.assertOptions(opts, exp)


class TestCreateMasterOptions(OptionsMixin, unittest.TestCase):
    def setUp(self):
        self.setUpOptions()

    def parse(self, *args):
        self.opts = runner.CreateMasterOptions()
        self.opts.parseOptions(args)
        return self.opts

    def defaults_and(self, **kwargs):
        defaults = {
            "force": False,
            "relocatable": False,
            "config": 'master.cfg',
            "db": 'sqlite:///state.sqlite',
            "basedir": os.getcwd(),
            "quiet": False,
            **{'no-logrotate': False, 'log-size': 10000000, 'log-count': 10},
        }
        unk_keys = set(kwargs.keys()) - set(defaults.keys())
        assert not unk_keys, f"invalid keys {unk_keys}"
        opts = defaults.copy()
        opts.update(kwargs)
        return opts

    def test_synopsis(self):
        opts = runner.CreateMasterOptions()
        self.assertIn('buildbot create-master', opts.getSynopsis())

    def test_defaults(self):
        opts = self.parse()
        exp = self.defaults_and()
        self.assertOptions(opts, exp)

    def test_db_quiet(self):
        opts = self.parse('-q')
        exp = self.defaults_and(quiet=True)
        self.assertOptions(opts, exp)

    def test_db_quiet_long(self):
        opts = self.parse('--quiet')
        exp = self.defaults_and(quiet=True)
        self.assertOptions(opts, exp)

    def test_force(self):
        opts = self.parse('-f')
        exp = self.defaults_and(force=True)
        self.assertOptions(opts, exp)

    def test_force_long(self):
        opts = self.parse('--force')
        exp = self.defaults_and(force=True)
        self.assertOptions(opts, exp)

    def test_relocatable(self):
        opts = self.parse('-r')
        exp = self.defaults_and(relocatable=True)
        self.assertOptions(opts, exp)

    def test_relocatable_long(self):
        opts = self.parse('--relocatable')
        exp = self.defaults_and(relocatable=True)
        self.assertOptions(opts, exp)

    def test_no_logrotate(self):
        opts = self.parse('-n')
        exp = self.defaults_and(**{'no-logrotate': True})
        self.assertOptions(opts, exp)

    def test_no_logrotate_long(self):
        opts = self.parse('--no-logrotate')
        exp = self.defaults_and(**{'no-logrotate': True})
        self.assertOptions(opts, exp)

    def test_config(self):
        opts = self.parse('-cxyz')
        exp = self.defaults_and(config='xyz')
        self.assertOptions(opts, exp)

    def test_config_long(self):
        opts = self.parse('--config=xyz')
        exp = self.defaults_and(config='xyz')
        self.assertOptions(opts, exp)

    def test_log_size(self):
        opts = self.parse('-s124')
        exp = self.defaults_and(**{'log-size': 124})
        self.assertOptions(opts, exp)

    def test_log_size_long(self):
        opts = self.parse('--log-size=124')
        exp = self.defaults_and(**{'log-size': 124})
        self.assertOptions(opts, exp)

    def test_log_size_noninteger(self):
        with self.assertRaises(usage.UsageError):
            self.parse('--log-size=1M')

    def test_log_count(self):
        opts = self.parse('-l124')
        exp = self.defaults_and(**{'log-count': 124})
        self.assertOptions(opts, exp)

    def test_log_count_long(self):
        opts = self.parse('--log-count=124')
        exp = self.defaults_and(**{'log-count': 124})
        self.assertOptions(opts, exp)

    def test_log_count_none(self):
        opts = self.parse('--log-count=None')
        exp = self.defaults_and(**{'log-count': None})
        self.assertOptions(opts, exp)

    def test_log_count_noninteger(self):
        with self.assertRaises(usage.UsageError):
            self.parse('--log-count=M')

    def test_db_long(self):
        opts = self.parse('--db=foo://bar')
        exp = self.defaults_and(db='foo://bar')
        self.assertOptions(opts, exp)

    def test_db_invalid(self):
        with self.assertRaisesRegex(usage.UsageError, "could not parse database URL 'inv_db_url'"):
            self.parse("--db=inv_db_url")

    def test_db_basedir(self):
        path = r'c:\foo\bar' if runtime.platformType == "win32" else '/foo/bar'
        opts = self.parse('-f', path)
        exp = self.defaults_and(force=True, basedir=path)
        self.assertOptions(opts, exp)


class BaseTestSimpleOptions(OptionsMixin):
    # tests for options with just --quiet and a usage message

    commandName: str | None = None
    optionsClass: type[usage.Options] | None = None

    def setUp(self):
        self.setUpOptions()

    def parse(self, *args):
        self.opts = self.optionsClass()
        self.opts.parseOptions(args)
        return self.opts

    def test_synopsis(self):
        opts = self.optionsClass()
        self.assertIn(f'buildbot {self.commandName}', opts.getSynopsis())

    def test_defaults(self):
        opts = self.parse()
        exp = {"quiet": False}
        self.assertOptions(opts, exp)

    def test_quiet(self):
        opts = self.parse('--quiet')
        exp = {"quiet": True}
        self.assertOptions(opts, exp)


class TestStopOptions(BaseTestSimpleOptions, unittest.TestCase):
    commandName = 'stop'
    optionsClass = runner.StopOptions


class TestResetartOptions(BaseTestSimpleOptions, unittest.TestCase):
    commandName = 'restart'
    optionsClass = runner.RestartOptions

    def test_nodaemon(self):
        opts = self.parse('--nodaemon')
        exp = {"nodaemon": True}
        self.assertOptions(opts, exp)


class TestStartOptions(BaseTestSimpleOptions, unittest.TestCase):
    commandName = 'start'
    optionsClass = runner.StartOptions

    def test_nodaemon(self):
        opts = self.parse('--nodaemon')
        exp = {"nodaemon": True}
        self.assertOptions(opts, exp)


class TestReconfigOptions(BaseTestSimpleOptions, unittest.TestCase):
    commandName = 'reconfig'
    optionsClass = runner.ReconfigOptions


class TestTryOptions(OptionsMixin, unittest.TestCase):
    def setUp(self):
        self.setUpOptions()

    def parse(self, *args):
        self.opts = runner.TryOptions()
        self.opts.parseOptions(args)
        return self.opts

    def defaults_and(self, **kwargs):
        defaults = {
            "connect": None,
            "host": None,
            "jobdir": None,
            "username": None,
            "master": None,
            "passwd": None,
            "who": None,
            "comment": None,
            "diff": None,
            "patchlevel": 0,
            "baserev": None,
            "vc": None,
            "branch": None,
            "repository": None,
            "topfile": None,
            "topdir": None,
            "wait": False,
            "dryrun": False,
            "quiet": False,
            "builders": [],
            "properties": {},
            "buildbotbin": 'buildbot',
        }
        # dashes make python syntax hard..
        defaults['get-builder-names'] = False
        if 'get_builder_names' in kwargs:
            kwargs['get-builder-names'] = kwargs['get_builder_names']
            del kwargs['get_builder_names']
        assert set(kwargs.keys()) <= set(defaults.keys()), "invalid keys"
        opts = defaults.copy()
        opts.update(kwargs)
        return opts

    def test_synopsis(self):
        opts = runner.TryOptions()
        self.assertIn('buildbot try', opts.getSynopsis())

    def test_defaults(self):
        opts = self.parse()
        exp = self.defaults_and()
        self.assertOptions(opts, exp)

    def test_properties(self):
        opts = self.parse('--properties=a=b')
        exp = self.defaults_and(properties={"a": 'b'})
        self.assertOptions(opts, exp)

    def test_properties_multiple_opts(self):
        opts = self.parse('--properties=X=1', '--properties=Y=2')
        exp = self.defaults_and(properties={"X": '1', "Y": '2'})
        self.assertOptions(opts, exp)

    def test_properties_equals(self):
        opts = self.parse('--properties=X=2+2=4')
        exp = self.defaults_and(properties={"X": '2+2=4'})
        self.assertOptions(opts, exp)

    def test_properties_commas(self):
        opts = self.parse('--properties=a=b,c=d')
        exp = self.defaults_and(properties={"a": 'b', "c": 'd'})
        self.assertOptions(opts, exp)

    def test_property(self):
        opts = self.parse('--property=a=b')
        exp = self.defaults_and(properties={"a": 'b'})
        self.assertOptions(opts, exp)

    def test_property_multiple_opts(self):
        opts = self.parse('--property=X=1', '--property=Y=2')
        exp = self.defaults_and(properties={"X": '1', "Y": '2'})
        self.assertOptions(opts, exp)

    def test_property_equals(self):
        opts = self.parse('--property=X=2+2=4')
        exp = self.defaults_and(properties={"X": '2+2=4'})
        self.assertOptions(opts, exp)

    def test_property_commas(self):
        opts = self.parse('--property=a=b,c=d')
        exp = self.defaults_and(properties={"a": 'b,c=d'})
        self.assertOptions(opts, exp)

    def test_property_and_properties(self):
        opts = self.parse('--property=X=1', '--properties=Y=2')
        exp = self.defaults_and(properties={"X": '1', "Y": '2'})
        self.assertOptions(opts, exp)

    def test_properties_builders_multiple(self):
        opts = self.parse('--builder=aa', '--builder=bb')
        exp = self.defaults_and(builders=['aa', 'bb'])
        self.assertOptions(opts, exp)

    def test_options_short(self):
        opts = self.parse(*'-n -q -c pb -u me -m mr:7 -w you -C comm -p 2 -b bb'.split())
        exp = self.defaults_and(
            dryrun=True,
            quiet=True,
            connect='pb',
            username='me',
            master='mr:7',
            who='you',
            comment='comm',
            patchlevel=2,
            builders=['bb'],
        )
        self.assertOptions(opts, exp)

    def test_options_long(self):
        opts = self.parse(
            *"""--wait --dryrun --get-builder-names --quiet --connect=pb
                --host=h --jobdir=j --username=u --master=m:1234 --passwd=p
                --who=w --comment=comm --diff=d --patchlevel=7 --baserev=br
                --vc=cvs --branch=br --repository=rep --builder=bl
                --properties=a=b --topfile=Makefile --topdir=.
                --buildbotbin=.virtualenvs/buildbot/bin/buildbot""".split()
        )
        exp = self.defaults_and(
            wait=True,
            dryrun=True,
            get_builder_names=True,
            quiet=True,
            connect='pb',
            host='h',
            jobdir='j',
            username='u',
            master='m:1234',
            passwd='p',
            who='w',
            comment='comm',
            diff='d',
            patchlevel=7,
            baserev='br',
            vc='cvs',
            branch='br',
            repository='rep',
            builders=['bl'],
            properties={"a": 'b'},
            topfile='Makefile',
            topdir='.',
            buildbotbin='.virtualenvs/buildbot/bin/buildbot',
        )
        self.assertOptions(opts, exp)

    def test_patchlevel_inval(self):
        with self.assertRaises(ValueError):
            self.parse('-p', 'a')

    def test_config_builders(self):
        self.options_file['try_builders'] = ['a', 'b']
        opts = self.parse()
        self.assertOptions(opts, {"builders": ['a', 'b']})

    def test_config_builders_override(self):
        self.options_file['try_builders'] = ['a', 'b']
        opts = self.parse('-b', 'd')  # overrides a, b
        self.assertOptions(opts, {"builders": ['d']})

    def test_config_old_names(self):
        self.options_file['try_masterstatus'] = 'ms'
        self.options_file['try_dir'] = 'td'
        self.options_file['try_password'] = 'pw'
        opts = self.parse()
        self.assertOptions(opts, {"master": 'ms', "jobdir": 'td', "passwd": 'pw'})

    def test_config_masterstatus(self):
        self.options_file['masterstatus'] = 'ms'
        opts = self.parse()
        self.assertOptions(opts, {"master": 'ms'})

    def test_config_masterstatus_override(self):
        self.options_file['masterstatus'] = 'ms'
        opts = self.parse('-m', 'mm')
        self.assertOptions(opts, {"master": 'mm'})

    def test_config_options(self):
        self.options_file.update({
            "try_connect": 'pb',
            "try_vc": 'cvs',
            "try_branch": 'br',
            "try_repository": 'rep',
            "try_topdir": '.',
            "try_topfile": 'Makefile',
            "try_host": 'h',
            "try_username": 'u',
            "try_jobdir": 'j',
            "try_password": 'p',
            "try_master": 'm:8',
            "try_who": 'w',
            "try_comment": 'comm',
            "try_quiet": 'y',
            "try_wait": 'y',
            "try_buildbotbin": '.virtualenvs/buildbot/bin/buildbot',
        })
        opts = self.parse()
        exp = self.defaults_and(
            wait=True,
            quiet=True,
            connect='pb',
            host='h',
            jobdir='j',
            username='u',
            master='m:8',
            passwd='p',
            who='w',
            comment='comm',
            vc='cvs',
            branch='br',
            repository='rep',
            topfile='Makefile',
            topdir='.',
            buildbotbin='.virtualenvs/buildbot/bin/buildbot',
        )
        self.assertOptions(opts, exp)

    def test_pb_withNoMaster(self):
        """
        When 'builbot try' is asked to connect via pb, but no master is
        specified, a usage error is raised.
        """
        with self.assertRaises(usage.UsageError):
            self.parse('--connect=pb')

    def test_pb_withInvalidMaster(self):
        """
        When 'buildbot try' is asked to connect via pb, but an invalid
        master is specified, a usage error is raised.
        """
        with self.assertRaises(usage.UsageError):
            self.parse('--connect=pb', '--master=foo')


class TestSendChangeOptions(OptionsMixin, unittest.TestCase):
    master_and_who = ['-m', 'm:1', '-W', 'w']

    def setUp(self):
        self.setUpOptions()
        self.getpass_response = 'typed-password'
        self.patch(getpass, 'getpass', lambda prompt: self.getpass_response)

    def parse(self, *args):
        self.opts = runner.SendChangeOptions()
        self.opts.parseOptions(args)
        return self.opts

    def test_synopsis(self):
        opts = runner.SendChangeOptions()
        self.assertIn('buildbot sendchange', opts.getSynopsis())

    def test_defaults(self):
        opts = self.parse('-m', 'm:1', '-W', 'me')
        exp = {
            "master": 'm:1',
            "auth": ('change', 'changepw'),
            "who": 'me',
            "vc": None,
            "repository": '',
            "project": '',
            "branch": None,
            "category": None,
            "revision": None,
            "revision_file": None,
            "property": None,
            "comments": '',
            "logfile": None,
            "when": None,
            "revlink": '',
            "encoding": 'utf8',
            "files": (),
        }
        self.assertOptions(opts, exp)

    def test_files(self):
        opts = self.parse(*[*self.master_and_who, 'a', 'b', 'c'])
        self.assertEqual(opts['files'], ('a', 'b', 'c'))

    def test_properties(self):
        opts = self.parse('--property', 'x:y', '--property', 'a:b', *self.master_and_who)
        self.assertEqual(opts['properties'], {"x": 'y', "a": 'b'})

    def test_properties_with_colon(self):
        opts = self.parse('--property', 'x:http://foo', *self.master_and_who)
        self.assertEqual(opts['properties'], {"x": 'http://foo'})

    def test_config_file(self):
        self.options_file['master'] = 'MMM:123'
        self.options_file['who'] = 'WWW'
        self.options_file['branch'] = 'BBB'
        self.options_file['category'] = 'CCC'
        self.options_file['vc'] = 'svn'
        opts = self.parse()
        exp = {"master": 'MMM:123', "who": 'WWW', "branch": 'BBB', "category": 'CCC', "vc": 'svn'}
        self.assertOptions(opts, exp)

    def test_short_args(self):
        opts = self.parse(
            *(
                '-m m:1 -a a:b -W W -R r -P p -b b -s git '
                + '-C c -r r -p pn:pv -c c -F f -w 123 -l l -e e'
            ).split()
        )
        exp = {
            "master": 'm:1',
            "auth": ('a', 'b'),
            "who": 'W',
            "repository": 'r',
            "project": 'p',
            "branch": 'b',
            "category": 'c',
            "revision": 'r',
            "vc": 'git',
            "properties": {"pn": 'pv'},
            "comments": 'c',
            "logfile": 'f',
            "when": 123.0,
            "revlink": 'l',
            "encoding": 'e',
        }
        self.assertOptions(opts, exp)

    def test_long_args(self):
        opts = self.parse(
            *(
                '--master m:1 --auth a:b --who w --repository r '
                + '--project p --branch b --category c --revision r --vc git '
                + '--property pn:pv --comments c --logfile f '
                + '--when 123 --revlink l --encoding e'
            ).split()
        )
        exp = {
            "master": 'm:1',
            "auth": ('a', 'b'),
            "who": 'w',
            "repository": 'r',
            "project": 'p',
            "branch": 'b',
            "category": 'c',
            "revision": 'r',
            "vc": 'git',
            "properties": {"pn": 'pv'},
            "comments": 'c',
            "logfile": 'f',
            "when": 123.0,
            "revlink": 'l',
            "encoding": 'e',
        }
        self.assertOptions(opts, exp)

    def test_revision_file(self):
        with open('revfile', "w", encoding='utf-8') as f:
            f.write('my-rev')
        self.addCleanup(lambda: os.unlink('revfile'))
        opts = self.parse('--revision_file', 'revfile', *self.master_and_who)
        self.assertOptions(opts, {"revision": 'my-rev'})

    def test_invalid_when(self):
        with self.assertRaises(usage.UsageError):
            self.parse('--when=foo', *self.master_and_who)

    def test_comments_overrides_logfile(self):
        opts = self.parse('--logfile', 'logs', '--comments', 'foo', *self.master_and_who)
        self.assertOptions(opts, {"comments": 'foo'})

    def test_logfile(self):
        with open('comments', "w", encoding='utf-8') as f:
            f.write('hi')
        self.addCleanup(lambda: os.unlink('comments'))
        opts = self.parse('--logfile', 'comments', *self.master_and_who)
        self.assertOptions(opts, {"comments": 'hi'})

    def test_logfile_stdin(self):
        stdin = mock.Mock()
        stdin.read = lambda: 'hi'
        self.patch(sys, 'stdin', stdin)
        opts = self.parse('--logfile', '-', *self.master_and_who)
        self.assertOptions(opts, {"comments": 'hi'})

    def test_auth_getpass(self):
        opts = self.parse('--auth=dustin', *self.master_and_who)
        self.assertOptions(opts, {"auth": ('dustin', 'typed-password')})

    def test_invalid_vcs(self):
        with self.assertRaises(usage.UsageError):
            self.parse('--vc=foo', *self.master_and_who)

    def test_invalid_master(self):
        with self.assertRaises(usage.UsageError):
            self.parse("--who=test", "-m foo")


class TestTryServerOptions(OptionsMixin, unittest.TestCase):
    def setUp(self):
        self.setUpOptions()

    def parse(self, *args):
        self.opts = runner.TryServerOptions()
        self.opts.parseOptions(args)
        return self.opts

    def test_synopsis(self):
        opts = runner.TryServerOptions()
        self.assertIn('buildbot tryserver', opts.getSynopsis())

    def test_defaults(self):
        with self.assertRaises(usage.UsageError):
            self.parse()

    def test_with_jobdir(self):
        opts = self.parse('--jobdir', 'xyz')
        exp = {"jobdir": 'xyz'}
        self.assertOptions(opts, exp)


class TestCheckConfigOptions(OptionsMixin, unittest.TestCase):
    def setUp(self):
        self.setUpOptions()

    def parse(self, *args):
        self.opts = runner.CheckConfigOptions()
        self.opts.parseOptions(args)
        return self.opts

    def test_synopsis(self):
        opts = runner.CheckConfigOptions()
        self.assertIn('buildbot checkconfig', opts.getSynopsis())

    def test_defaults(self):
        opts = self.parse()
        exp = {"quiet": False}
        self.assertOptions(opts, exp)

    def test_configfile(self):
        opts = self.parse('foo.cfg')
        exp = {"quiet": False, "configFile": 'foo.cfg'}
        self.assertOptions(opts, exp)

    def test_quiet(self):
        opts = self.parse('-q')
        exp = {"quiet": True}
        self.assertOptions(opts, exp)


class TestUserOptions(OptionsMixin, unittest.TestCase):
    # mandatory arguments
    extra_args = ['--master', 'a:1', '--username', 'u', '--passwd', 'p']

    def setUp(self):
        self.setUpOptions()

    def parse(self, *args):
        self.opts = runner.UserOptions()
        self.opts.parseOptions(args)
        return self.opts

    def test_defaults(self):
        with self.assertRaises(usage.UsageError):
            self.parse()

    def test_synopsis(self):
        opts = runner.UserOptions()
        self.assertIn('buildbot user', opts.getSynopsis())

    def test_master(self):
        opts = self.parse(
            "--master", "abcd:1234", '--op=get', '--ids=x', '--username=u', '--passwd=p'
        )
        self.assertOptions(opts, {"master": 'abcd:1234'})

    def test_ids(self):
        opts = self.parse("--ids", "id1,id2,id3", '--op', 'get', *self.extra_args)
        self.assertEqual(opts['ids'], ['id1', 'id2', 'id3'])

    def test_info(self):
        opts = self.parse(
            "--info", "git=Tyler Durden <tyler@mayhem.net>", '--op', 'add', *self.extra_args
        )
        self.assertEqual(opts['info'], [{"git": 'Tyler Durden <tyler@mayhem.net>'}])

    def test_info_only_id(self):
        opts = self.parse("--info", "tdurden", '--op', 'update', *self.extra_args)
        self.assertEqual(opts['info'], [{"identifier": 'tdurden'}])

    def test_info_with_id(self):
        opts = self.parse("--info", "tdurden:svn=marla", '--op', 'update', *self.extra_args)
        self.assertEqual(opts['info'], [{"identifier": 'tdurden', "svn": 'marla'}])

    def test_info_multiple(self):
        opts = self.parse(
            "--info",
            "git=Tyler Durden <tyler@mayhem.net>",
            "--info",
            "git=Narrator <narrator@mayhem.net>",
            '--op',
            'add',
            *self.extra_args,
        )
        self.assertEqual(
            opts['info'],
            [{"git": 'Tyler Durden <tyler@mayhem.net>'}, {"git": 'Narrator <narrator@mayhem.net>'}],
        )

    def test_config_user_params(self):
        self.options_file['user_master'] = 'mm:99'
        self.options_file['user_username'] = 'un'
        self.options_file['user_passwd'] = 'pw'
        opts = self.parse('--op', 'get', '--ids', 'x')
        self.assertOptions(opts, {"master": 'mm:99', "username": 'un', "passwd": 'pw'})

    def test_config_master(self):
        self.options_file['master'] = 'mm:99'
        opts = self.parse('--op', 'get', '--ids', 'x', '--username=u', '--passwd=p')
        self.assertOptions(opts, {"master": 'mm:99'})

    def test_config_master_override(self):
        self.options_file['master'] = 'not seen'
        self.options_file['user_master'] = 'mm:99'
        opts = self.parse('--op', 'get', '--ids', 'x', '--username=u', '--passwd=p')
        self.assertOptions(opts, {"master": 'mm:99'})

    def test_invalid_info(self):
        with self.assertRaises(usage.UsageError):
            self.parse("--info", "foo=bar", '--op', 'add', *self.extra_args)

    def test_no_master(self):
        with self.assertRaises(usage.UsageError):
            self.parse('-op=foo')

    def test_invalid_master(self):
        with self.assertRaises(usage.UsageError):
            self.parse('-m', 'foo')

    def test_no_operation(self):
        with self.assertRaises(usage.UsageError):
            self.parse('-m', 'a:1')

    def test_bad_operation(self):
        with self.assertRaises(usage.UsageError):
            self.parse('-m', 'a:1', '--op=mayhem')

    def test_no_username(self):
        with self.assertRaises(usage.UsageError):
            self.parse('-m', 'a:1', '--op=add')

    def test_no_password(self):
        with self.assertRaises(usage.UsageError):
            self.parse('--op=add', '-m', 'a:1', '-u', 'tdurden')

    def test_invalid_bb_username(self):
        with self.assertRaises(usage.UsageError):
            self.parse('--op=add', '--bb_username=tdurden', *self.extra_args)

    def test_invalid_bb_password(self):
        with self.assertRaises(usage.UsageError):
            self.parse('--op=add', '--bb_password=marla', *self.extra_args)

    def test_update_no_bb_username(self):
        with self.assertRaises(usage.UsageError):
            self.parse('--op=update', '--bb_password=marla', *self.extra_args)

    def test_update_no_bb_password(self):
        with self.assertRaises(usage.UsageError):
            self.parse('--op=update', '--bb_username=tdurden', *self.extra_args)

    def test_no_ids_info(self):
        with self.assertRaises(usage.UsageError):
            self.parse('--op=add', *self.extra_args)

    def test_ids_with_add(self):
        with self.assertRaises(usage.UsageError):
            self.parse('--op=add', '--ids=id1', *self.extra_args)

    def test_ids_with_update(self):
        with self.assertRaises(usage.UsageError):
            self.parse('--op=update', '--ids=id1', *self.extra_args)

    def test_no_ids_found_update(self):
        with self.assertRaises(usage.UsageError):
            self.parse("--op=update", "--info=svn=x", *self.extra_args)

    def test_id_with_add(self):
        with self.assertRaises(usage.UsageError):
            self.parse("--op=add", "--info=id:x", *self.extra_args)

    def test_info_with_remove(self):
        with self.assertRaises(usage.UsageError):
            self.parse('--op=remove', '--info=x=v', *self.extra_args)

    def test_info_with_get(self):
        with self.assertRaises(usage.UsageError):
            self.parse('--op=get', '--info=x=v', *self.extra_args)


class TestOptions(OptionsMixin, misc.StdoutAssertionsMixin, unittest.TestCase):
    def setUp(self):
        self.setUpOptions()
        self.setUpStdoutAssertions()

    def parse(self, *args):
        self.opts = runner.Options()
        self.opts.parseOptions(args)
        return self.opts

    def test_defaults(self):
        with self.assertRaises(usage.UsageError):
            self.parse()

    def test_version(self):
        try:
            self.parse('--version')
        except SystemExit as e:
            self.assertEqual(e.args[0], 0)
        self.assertInStdout('Buildbot version:')

    def test_verbose(self):
        self.patch(log, 'startLogging', mock.Mock())
        with self.assertRaises(usage.UsageError):
            self.parse("--verbose")
        log.startLogging.assert_called_once_with(sys.stderr)


class TestRun(unittest.TestCase):
    class MySubCommand(usage.Options):
        subcommandFunction = 'buildbot.test.unit.scripts.test_runner.subcommandFunction'
        optFlags = [['loud', 'l', 'be noisy']]

        def postOptions(self):
            if self['loud']:
                raise usage.UsageError('THIS IS ME BEING LOUD')

    def setUp(self):
        # patch our subcommand in
        self.patch(runner.Options, 'subCommands', [['my', None, self.MySubCommand, 'my, my']])

        # and patch in the callback for it
        global subcommandFunction
        subcommandFunction = mock.Mock(name='subcommandFunction', return_value=3)

    def test_run_good(self):
        self.patch(sys, 'argv', ['buildbot', 'my'])
        try:
            runner.run()
        except SystemExit as e:
            self.assertEqual(e.args[0], 3)
        else:
            self.fail("didn't exit")

    def test_run_bad(self):
        self.patch(sys, 'argv', ['buildbot', 'my', '-l'])
        stdout = StringIO()
        self.patch(sys, 'stdout', stdout)
        try:
            runner.run()
        except SystemExit as e:
            self.assertEqual(e.args[0], 1)
        else:
            self.fail("didn't exit")
        self.assertIn('THIS IS ME', stdout.getvalue())



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import Sequence

from buildbot.util import unicode2bytes
from buildbot.util.sautils import hash_columns


class Row:
    """
    Parent class for row classes, which are used to specify test data for
    database-related tests.

    @cvar table: the table name

    @cvar id_column: specify a column that should be assigned an
    auto-incremented id.  Auto-assigned id's begin at 1000, so any explicitly
    specified ID's should be less than 1000.

    @cvar hashedColumns: a tuple of hash column and source columns designating
    a hash to work around MySQL's inability to do indexing.

    @ivar values: the values to be inserted into this row
    """

    id_column: tuple[()] | str = ()

    hashedColumns: Sequence[tuple[str, Sequence[str]]] = ()
    # Columns that content is represented as sa.Binary-like type in DB model.
    # They value is bytestring (in contrast to text-like columns, which are
    # unicode).
    binary_columns: Sequence[str] = ()

    _next_id = None

    def __init__(self, **kwargs):
        if self.__init__.__func__ is Row.__init__:
            raise RuntimeError(
                'Row.__init__ must be overridden to supply default values for columns'
            )

        self.values = kwargs.copy()
        if self.id_column:
            if self.values[self.id_column] is None:
                self.values[self.id_column] = self.nextId()
        # Binary columns stores either (compressed) binary data or encoded
        # with utf-8 unicode string. We assume that Row constructor receives
        # only unicode strings and encode them to utf-8 here.
        # At this moment there is only one such column: logchunks.contents,
        # which stores either utf-8 encoded string, or gzip-compressed
        # utf-8 encoded string.
        for col in self.binary_columns:
            self.values[col] = unicode2bytes(self.values[col])
        # calculate any necessary hashes
        for hash_col, src_cols in self.hashedColumns:
            self.values[hash_col] = hash_columns(*(self.values[c] for c in src_cols))

        # make the values appear as attributes
        self.__dict__.update(self.values)

    def __repr__(self):
        values_str = ''.join(f'{k}={v!r}, ' for k, v in self.values.items())
        return f'{self.__class__.__name__}({values_str})'

    @staticmethod
    def nextId():
        id = Row._next_id if Row._next_id is not None else 1
        Row._next_id = id + 1
        return id
